# Modules dependencies
import pandas as pd
import numpy as np
import os
from scipy.spatial.distance import cdist

class FlattenerException(Exception):
    pass

class Flattener():
    """
    This class is design to create flattening coordinate operations over centroids
    coordinate operations. It works mainly using flattening aproachs.
    """

    def __init__(
        self,
        grid_data,
        grid_model_height=5,
        xc_coordinates_label="XC",
        yc_coordinates_label="YC",
        zc_coordinates_label="elevation",
        print_info=False,
        approach='mid wall'):

        self.grid_data = grid_data
        self.grid_model_height = grid_model_height
        self.xc_coordinates_label = xc_coordinates_label
        self.yc_coordinates_label = yc_coordinates_label
        self.zc_coordinates_label = zc_coordinates_label
        self.coordinates_labels = [
            xc_coordinates_label,
            yc_coordinates_label,
            zc_coordinates_label]
        self.print_info = print_info

        available_approaches = [
            "mid wall", "hanging wall", "foot wall", "mid wall continous"]
        if not approach in available_approaches:
            raise NotImplemented(
                f"The coordinate flatenning of type {approach} isn't implemented"
                )
        self.approach = approach

        # A in memory storing of the computed displacements
        self.displacements = dict()

        self.grid_data_extractor()


    def grid_data_extractor(self):
        # Group per x-y coordinates
        xy_columns_labels = [self.xc_coordinates_label, self.yc_coordinates_label]
        xy_data_grouped = self.grid_data.groupby(xy_columns_labels)

        # Some data to retrieve from xy grouped coordinates
        ## Max z coordinates
        xy_data_grouped_elevation_max = xy_data_grouped[self.zc_coordinates_label].max()
        xy_data_grouped_elevation_min = xy_data_grouped[self.zc_coordinates_label].min()
        ## z coordinate ranges on each xy group
        xy_elevation_ranges = xy_data_grouped_elevation_max - xy_data_grouped_elevation_min
        ## Maximum range and x-y location
        x_coord_max, y_coord_max = xy_elevation_ranges.idxmax()
        max_column_height = xy_elevation_ranges.max()
        ## Cuboids data at maximal range xy position
        max_column_data = self.grid_data.loc[(
            self.grid_data[self.xc_coordinates_label] == x_coord_max) \
            & (self.grid_data[self.yc_coordinates_label] == y_coord_max)]
        ## Data along the maximum range column
        max_column_max_height = max_column_data[self.zc_coordinates_label].max()
        max_column_min_height =  max_column_data[self.zc_coordinates_label].min()
        max_column_cuboids = int(max_column_height/self.grid_model_height) + 1

        if self.print_info:
            # A brief resume and human readable information print
            if max_column_height % self.grid_model_height != 0:
                raise FlattenerException("Unexpected behavior check grid_model_height")
            else:
                info_message_total_height = "Largest column has {} meters height between centers"
                info_message_total_cuboids = "Largest column has {} cuboids"
                info_message_largest_column_range = "Largest ranges from {} to {} in z axis"
                info_max_column_xy_coordinates = "Largest column is located at x:{} y:{}"
                print(info_message_total_height.format(max_column_height))
                print(info_message_total_cuboids.format(max_column_cuboids))
                print(info_max_column_xy_coordinates.format(x_coord_max, y_coord_max))
                print(info_message_largest_column_range.format(max_column_min_height, max_column_max_height))

        # Store grid data in object attributes
        self.xy_data_grouped= xy_data_grouped
        self.xy_data_grouped_elevation_max = xy_data_grouped_elevation_max
        self.xy_data_grouped_elevation_min = xy_data_grouped_elevation_min
        self.max_column_height = max_column_height
        self.max_column_cuboids = max_column_cuboids
        self.max_column_max_height = max_column_max_height
        self.max_column_min_height = max_column_min_height
        self.x_coord_max = x_coord_max
        self.y_coord_max = y_coord_max
        self.max_column_data = max_column_data


    def mid_flattening_displacements(self):
        """
        A set of instructions to perform a flattening coordinate change. It returns
        a dataframe of displacements, indexed as the input grid_data parameter, which
        value is the total displacement required for each indexed coordinate center
        to be transformed to a flattened coordinate system.

        Return a Pandas.Series of the indexed displacements to mid_flattening coordinates.
        """

        # Vertical flattening operations
        ## Vertical slots of coordinate flattening operation
        zc_slots_new_coordinates = np.linspace(
            start=self.max_column_min_height,
            stop=self.max_column_max_height,
            num=self.max_column_cuboids)

        ## Posible vertical coordinates, the key is the number of cuboids
        ## the value is the transformed z coordinates
        possible_coordinates = {}
        for n_cuboids in range(self.max_column_cuboids+1):
            positions = ((np.arange(n_cuboids)+1)/(n_cuboids+1)*self.max_column_cuboids).astype(int)
            possible_coordinates[n_cuboids] = zc_slots_new_coordinates[positions]

        ## Assamble new coordinate columns of z coordinates
        new_coordinates_column = pd.Series(index=self.grid_data.index, dtype='float64')
        for indices, data in self.xy_data_grouped:
            new_coordinates = possible_coordinates[len(data)]
            sorted_data = data[self.zc_coordinates_label].sort_values()
            new_coordinates_column.loc[sorted_data.index] = new_coordinates

        ## Return displacements
        displacements = new_coordinates_column - self.grid_data.loc[:, self.zc_coordinates_label]
        displacements.name = 'displacements'

        return displacements

    def mid_continous_flattening_displacements(self):

        xy_elevation_ranges = \
            self.xy_data_grouped_elevation_max - self.xy_data_grouped_elevation_min + self.grid_model_height
        xy_elevation_ranges = xy_elevation_ranges.rename('range')

        xy_data_grouped_elevation_min = self.xy_data_grouped_elevation_min.rename('min') -  self.grid_model_height/2

        xy_range_min = pd.concat(
            [xy_elevation_ranges, xy_data_grouped_elevation_min], axis=1)
        xy_range_min = xy_range_min.reset_index()

        grid_data = self.grid_data.loc[:, self.coordinates_labels]

        xy_columns_labels = [self.xc_coordinates_label, self.yc_coordinates_label]
        grid_data = grid_data.merge(xy_range_min, on=xy_columns_labels, how='left')
        grid_data.index = self.grid_data.index

        z_proportion = f"({self.zc_coordinates_label} - min ) / range"
        z_proportion_data = grid_data.eval(z_proportion).fillna(0)

        grid_data['z_flattening'] = \
            z_proportion_data * (self.max_column_height + self.grid_model_height) + self.max_column_min_height - self.grid_model_height/2


        displacements = grid_data.eval(
            f'z_flattening - {self.zc_coordinates_label}')

        displacements = displacements.rename('displacements')
        return displacements

    def wall_coordinates_displacements(self):
        """
        A set of instructions to perform a flattening coordinate change to foot or hanging wall.
        It returns a dataframe of displacements, indexed as the input grid_data parameter, which
        value is the total displacement required for each indexed coordinate center
        to be transformed to a flattened coordinate system.

        Returns a Pandas.Series of the indexed displacements to the given coordinates types.
        """

        ## Compute displacements
        if self.approach == 'hanging wall':
            displacements = self.max_column_max_height - self.xy_data_grouped_elevation_max
        elif self.approach == 'foot wall':
            displacements = self.max_column_min_height - self.xy_data_grouped_elevation_min

        displacements = displacements.reset_index()
        displacements.columns = [self.xc_coordinates_label, self.yc_coordinates_label, 'displacements']

        displacements = self.grid_data.merge(displacements, on=[self.xc_coordinates_label, self.yc_coordinates_label], how='left')
        return displacements.loc[:, 'displacements']


    def compute_displacements(self):
        """
        Computes zc_coordinates displacements to a flattening position.
        Flattening types available are
            ["mid wall", "hanging wall", "foot wall", "mid wall continous"].
        """

        if self.print_info:
            print(f"Computing displacements type {self.approach}")

        already_computed_displacements = self.displacements.get(self.approach)
        if isinstance(already_computed_displacements, pd.Series):
            return already_computed_displacements

        if self.approach == "mid wall":
            displacements = self.mid_flattening_displacements()
        elif self.approach == "mid wall continous":
            displacements = self.mid_continous_flattening_displacements()
        else:
            displacements = self.wall_coordinates_displacements()


        self.displacements[self.approach] = displacements

        return displacements


    def nn_displacements(self, coordinates,
        x_coordinates_label=None, y_coordinates_label=None, z_coordinates_label=None):
        """
        Computes the nearest neighbor displacements. This mean
        a displacements vector of the same length as coordinates,
        which tells how to "move" to flattened coordinates.
        """

        grid_data = self.grid_data[self.coordinates_labels]

        coordinates_labels = self.coordinates_labels
        if x_coordinates_label:
            coordinates_labels[0] = x_coordinates_label
        if y_coordinates_label:
            coordinates_labels[1] = y_coordinates_label
        if y_coordinates_label:
            coordinates_labels[2] = z_coordinates_label

        coordinates = coordinates[coordinates_labels]

        distance_matrix = pd.DataFrame(cdist(
            grid_data, coordinates, 'euclidean'))

        nearest_neighbors = distance_matrix.idxmin()

        displacements = self.compute_displacements()

        nn_displacements = displacements.iloc[nearest_neighbors.values]
        nn_displacements.index = coordinates.index

        return nn_displacements
