
import rmsp
import pandas as pd


def search_per_row_constructor(row, print_info=False):
    """ Construct a rmsp.Search object for each row contained
    on the search parameters sheet of a parameters file"""

    searches = []
    search_range_x = row.loc['SDIST1']
    search_range_y = row.loc['SDIST2']
    search_range_z = row.loc['SDIST3']
    search_range = [search_range_x, search_range_y, search_range_z]

    search_angle_a = row.loc['SANGLE1']
    search_angle_b = row.loc['SANGLE2']
    search_angle_c = row.loc['SANGLE3']

    search_angle = [search_angle_a, search_angle_b, search_angle_c]


    max_comps_per_dhid =  row.loc['MAXKEY']
    max_comps_per_oct =  row.loc['MAXPEROC']

    i = 1
    min_comps =  row.loc[f"MINNUM{i}"]
    max_comps =  row.loc[f"MAXNUM{i}"]

    searches.append(
        rmsp.Search(
            angles=search_angle,
            ranges=search_range,
            min_comps=min_comps,
            max_comps=max_comps,
            max_comps_per_dhid=max_comps_per_dhid,
            max_comps_per_oct=max_comps_per_oct)
    )

    for i in [2,3]:
        min_comps =  row.loc[f"MINNUM{i}"]
        max_comps =  row.loc[f"MAXNUM{i}"]
        volume_factor = row.loc[f"SVOLFAC{i}"]

        is_valid_min_comps = isinstance(min_comps,int)
        is_valid_max_comps = isinstance(max_comps,int)
        is_valid_volume_factor = isinstance(volume_factor, (int, float))

        valid_new_search = all(
            [is_valid_min_comps, is_valid_max_comps, is_valid_volume_factor])

        if valid_new_search:
            searches.append(
                rmsp.Search(
                    angles=search_angle,
                    ranges=[
                        search_range_x * volume_factor,
                        search_range_y * volume_factor,
                        search_range_z * volume_factor],
                    min_comps=min_comps,
                    max_comps=max_comps,
                    max_comps_per_dhid=max_comps_per_dhid,
                    max_comps_per_oct=max_comps_per_oct)
            )

    return searches


def variomodel_per_row_constructor(row,  print_info=False):
    """
    A method to retrieve rmsp.Searches and rmsp.VarioModel objects
    from Datamine like exports. These exports needs to be stored inside
    an Excel file and include the columns "DOMAIN" and "VAR"
    """

    angle_a = row['VANGLE1']
    angle_b = row['VANGLE2']
    angle_c = row['VANGLE3']
    angles = [angle_a, angle_b, angle_c]

    nugget = row['NUGGET']

    # Iterate over struct columns retrieving ranges, variances and shapes
    i = 1
    range_vars = lambda:  [f"ST{i}PAR1", f"ST{i}PAR2", f"ST{i}PAR3", f"ST{i}PAR4", f"ST{i}"]
    columns_expected = range_vars()
    columns_exists = bool(set(columns_expected).intersection(set(row.index)))
    ranges = []
    var_contribs = []
    shapes = []
    num_struct = 0
    shapes_translation = {1: 'spherical', 2: 'exponential'}
    while columns_exists:


        range_x = row[columns_expected[0]]
        range_y = row[columns_expected[1]]
        range_z = row[columns_expected[2]]
        struct_ranges = [range_x, range_y, range_z]

        struct_var_contribs = row[columns_expected[3]]
        struct_shape = row[columns_expected[4]]

        # Validation of struct data
        range_x_is_valid = isinstance(range_x, float) and range_x is not 0
        range_y_is_valid = isinstance(range_y, float) and range_y is not 0
        range_z_is_valid = isinstance(range_z, float) and range_z is not 0
        var_contribs_is_valid = isinstance(struct_var_contribs, float)
        shape_is_valid = isinstance(struct_shape, int) and struct_shape in [1, 2]

        struct_is_valid = all(
            [range_x_is_valid, range_y_is_valid, range_z_is_valid,
             var_contribs_is_valid, shape_is_valid])

        if not struct_is_valid:
            break

        # Assambly vario dict to construct rmsp.VarioModel
        num_struct = num_struct +1

        ranges.append(struct_ranges)
        var_contribs.append(struct_var_contribs)
        shapes.append(shapes_translation[struct_shape])

        i = i+1
        columns_expected = range_vars()
        columns_exists = bool(set(columns_expected).intersection(set(row.index)))


    angles = [angles]*num_struct


    vario_dict = {
        'num_struct': num_struct ,
        'nugget': nugget,
        'shapes': shapes,
        'var_contribs': var_contribs,
        'angles':angles,
        'ranges': ranges,
        }


    vario_model = rmsp.VarioModel(vario_dict)

    if print_info:
        print("\t variomod ranges {}".format(ranges))
        print("\t variomod angles {}".format(angles))
        print("\t variomod nugget {}".format(nugget))
        print("\t variomod var_contribs {}".format(var_contribs))
        print("\t variomod shapes {}".format(shapes))
        print("\t variomod number of structs {}".format(num_struct))
        print(vario_model)

    return vario_model



def retrieve_parameters_from_file(parameter_file_name):
    """
    A method to retrieve rmsp.Searches and rmsp.VarioModel objects
    from Datamine like exports. These exports needs to be stored inside
    an Excel file and include the columns "DOMAIN" and "VAR"
    """

    search_parameters = pd.read_excel(
            parameter_file_name,
            sheet_name="search parameters")

    search_parameters['rmsp.Search'] = search_parameters.apply(search_per_row_constructor,axis=1)

    search_parameters_dict = {
        estimator: {
            domain: {
                var: search_per_estimator_and_domain_and_var['rmsp.Search'].iloc[0]
                for var, search_per_estimator_and_domain_and_var in search_per_estimator_and_domain.groupby('VAR')
            }
            for domain, search_per_estimator_and_domain in search_per_estimator.groupby('DOMAIN')
        }
        for estimator, search_per_estimator in search_parameters.groupby('ESTI')
    }

    variomod_parameters = pd.read_excel(
        parameter_file_name,
        sheet_name="vario models parameters")

    variomod_parameters['rmsp.VarioMod'] = variomod_parameters.apply(variomodel_per_row_constructor, axis=1)

    variomod_parameters_dict = {
        domain: {
            var: variomod_per_domain_and_var['rmsp.VarioMod'].iloc[0]
            for var, variomod_per_domain_and_var in variomod_per_domain.groupby('VAR')
        }
        for domain, variomod_per_domain in variomod_parameters.groupby('DOMAIN')
    }



    return search_parameters_dict, variomod_parameters_dict
