# Modules dependencies
import pandas as pd
import numpy as np
import time
from scipy.spatial.distance import cdist

class EquivalentGridException(Exception):
    pass

def equivalent_grid_computation(
    centroids_data,
    composites_data,
    x_centroid_coordinates_label="XC",
    y_centroid_coordinates_label="YC",
    z_centroid_coordinates_label="ZC",
    x_composites_coordinates_label="X",
    y_composites_coordinates_label="Y",
    z_composites_coordinates_label="Z",
    k_nearest_composites=3,
    equivalent_grid_factor=2**(1/2),
    print_info=False,
):

    centroid_coordinates_labels = [
        x_centroid_coordinates_label,
        y_centroid_coordinates_label,
        z_centroid_coordinates_label
    ]
    centroids_data = centroids_data.loc[:, centroid_coordinates_labels].values
    total_centroids = len(centroids_data)
    if print_info:
        print(f"\t Received {total_centroids} centroids to operate with")


    composites_coordinaltes_labels = [
        x_composites_coordinates_label,
        y_composites_coordinates_label,
        z_composites_coordinates_label]
    composites_data = composites_data.loc[:, composites_coordinaltes_labels].values
    total_composites = len(composites_data)
    if print_info:
        print(f"\t Received {total_composites} composites to operate with")

    iterator_size = min(150, max(50, int(total_composites/20)))
    iterator_steps = int(np.ceil(total_composites/iterator_size)-1)
    equivalent_grids = np.full(
        len(centroids_data),
        np.inf)

    if print_info:
        print(f"\t Processing equivalent grids")

    start_time = time.time()
    distances = np.empty(shape=(total_centroids, 0))
    for step, i in enumerate(range(0, total_composites, iterator_size)):

        distances_computed = cdist(
            centroids_data,
            composites_data[i:i+iterator_size,:]
            )

        distances = np.sort(np.hstack([
            distances,
            distances_computed
        ]))[:,0:k_nearest_composites]


        step_computed_equivalent_grids = equivalent_grid_factor*np.sort(
            distances)[:,0:k_nearest_composites].mean(axis=1)

        improved_estimations = step_computed_equivalent_grids < equivalent_grids

        equivalent_grids[improved_estimations] = step_computed_equivalent_grids[improved_estimations]

        if print_info:
            elapsed_time = time.time() - start_time
            mean_time = elapsed_time/(step+1)
            estimated_time = mean_time*(iterator_steps)
            print(f"\t\t Computation step {step} of {iterator_steps} ")
            print(f"\t\t Elapsed time {elapsed_time:.2f} s.")
            print(f"\t\t Mean time per step so far {mean_time:.2f} s.")
            print(f"\t\t Estimated time ahead {(estimated_time/60):.2f} min.")

    return equivalent_grids
