import pandas as pd
import numpy as np
import rmsp
import itertools
import random
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde

pd.options.mode.chained_assignment = None

rmsp.activate()

class RmspMovingPanelException(Exception):
    pass

class RmspMovingPanelValidator:
    collected_data = pd.DataFrame(
        columns=[
            'index',
            'blocks_average',
            'blocks_filled',
            'composites_in_panel',
            'weighted_measurement',
            'iteration',
            'variable',
            'domain',
            'displacements',
            'xc',
            'yc',
            'zc'])

    def __init__(
        self,
        grid_data,
        composites,
        iterations=12,
        panel_usize=50.0,
        panel_vsize=50.0,
        panel_zsize=50.0,
        print_info=False,
        domains=None,
        domain_column=None,
        variables=['CUT'],
        ):

        if not isinstance(grid_data, rmsp.GridData):
            error_msg = 'grid_data need to be a rmsp.GridData instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.grid_data = grid_data
            self.grid_def = self.deduce_grid_def(self.grid_data) # Panel griddef

        if not isinstance(composites, rmsp.PointData):
            error_msg = 'composites need to be a rmsp.GridData instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.composites = composites

        if not isinstance(iterations, int):
            error_msg = 'iterations need to be a int instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.iterations = iterations

        if not isinstance(panel_usize, float):
            error_msg = 'panel_usize need to be a float instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.panel_usize = panel_usize

        if not isinstance(panel_vsize, float):
            error_msg = 'panel_vsize need to be a float instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.panel_vsize = panel_vsize

        if not isinstance(panel_zsize, float):
            error_msg = 'panel_zsize need to be a float instance'
            raise RmspMovingPanelException(error_msg)
        else:
            self.panel_zsize = panel_zsize

        if  domain_column == None:
            self.domain_colum = None
        elif not isinstance(domain_column, str):
            error_msg = "domain_column need to be a string or None"
            raise RmspMovingPanelException(error_msg)
        elif not domain_column in self.grid_data.columns:
            error_msg = "domain_column need to be a column of grid_data"
            raise RmspMovingPanelException(error_msg)
        elif not domain_column in self.composites.columns:
            error_msg = "domain_column need to be a column of composites"
            raise RmspMovingPanelException(error_msg)
        else:
            self.domain_column = domain_column

        if  domains == None:
            if domain_column != None:
                self.domains = self.grid_data[domain_column].unique()
            else:
                self.domains = [None]
        elif not isinstance(domains, list):
            error_msg = 'domains need to be a list instance or None'
            raise RmspMovingPanelException(error_msg)
        else:
            self.domains = domains

        if not isinstance(variables, list):
            error_msg = "variables need to be a list instance"
            raise RmspMovingPanelException(error_msg)
        else:
            for variable in variables:
                if not isinstance(variable, str):
                    error_msg = "all variables in variables need to be a string instance"
                    raise RmspMovingPanelException(error_msg)
                elif not variable in self.composites.columns:
                    error_msg = f"variable {variable} needs to be a column in composites"
                    raise RmspMovingPanelException(error_msg)
                elif not variable in self.grid_data.columns:
                    error_msg = f"variable {variable} needs to be a column in grid_data"
                    raise RmspMovingPanelException(error_msg)
            self.variables = variables

        self.print_info = print_info


    def deduce_grid_def(self, grid_data):
        if not self.grid_data.griddef is None:
            return self.grid_data.griddef
        else:
            deduced_grid_def = {}
            coordinates = grid_data.coords()
            coordinates_labels = ['XC', 'YC', 'ZC']
            for coordinate_label, coordinate_column in zip(
                coordinates_labels, coordinates):
                coords = np.sort(
                    np.unique(
                        coordinate_column.unique()))
                min_coords = min(coords)
                max_coords = max(coords)
                total_elements = len(coords)
                size = (max_coords-min_coords) / (total_elements -1)

                expected_coordinates = np.linspace(
                    min_coords,
                    max_coords,
                    total_elements)

                if not np.all(expected_coordinates == coords):
                    print(f'Cant deduce griddef for {coordinate_column}')

                deduced_grid_def[coordinate_label] = {
                    'min': min_coords,
                    'max': max_coords,
                    'n': total_elements,
                    'size': size}

            grid_def_parameters = dict(
                xcmin = deduced_grid_def['XC']['min'],
                nu = deduced_grid_def['XC']['n'],
                usize = deduced_grid_def['XC']['size'],
                ycmin = deduced_grid_def['YC']['min'],
                nv = deduced_grid_def['YC']['n'],
                vsize = deduced_grid_def['YC']['size'],
                zcmin = deduced_grid_def['ZC']['min'],
                nz = deduced_grid_def['ZC']['n'],
                zsize = deduced_grid_def['ZC']['size'],
                azimuth_deg = 0)

            model_grid_def = rmsp.GridDef(
                **grid_def_parameters,
                blockindices='full')

            if self.print_info:
                print(f'The deduced GridDef parameters is: {model_grid_def}')

            return model_grid_def

    def all_posible_movements(self):
        model_grid_def = self.grid_def
        u_range =  self.panel_usize/2 - model_grid_def.usize
        v_range =  self.panel_vsize/2 - model_grid_def.vsize
        z_range =  self.panel_zsize/2 - model_grid_def.zsize

        u_displacements = np.arange(-u_range+1, u_range)
        v_displacements = np.arange(-v_range+1, v_range)
        z_displacements = np.arange(-z_range+1, z_range)

        all_possible_displacements = list(
            itertools.product(
                u_displacements,
                v_displacements,
                z_displacements))

        return all_possible_displacements

    def collect_panels_data(self, trajectory=None):

        self.collected_data = self.collected_data[0:0]

        model_grid_def = self.grid_def
        all_possible_displacements = self.all_posible_movements()

        if not trajectory:
            displacements = random.sample(
                all_possible_displacements,
                self.iterations)
        else:
            displacements = trajectory
            self.iterations = len(displacements)

        # Make displacement to the third quadrant only
        fixed_displacements = []
        for x, y, z in displacements:
            if x > 0:
                u = x-np.ceil(x/self.panel_usize)*self.panel_usize
            else:
                u=x
            if y > 0:
                v = y-np.ceil(y/self.panel_vsize)*self.panel_vsize
            else:
                v=y
            fixed_displacements.append(
                [u, v, z]
            )
        displacements = fixed_displacements

        if model_grid_def.azimuth_deg != 0:
            theta = 90 + model_grid_def.azimuth_deg
            if self.print_info:
                print("The deduced GridDef is rotated, "
                    "modifying displacements in a rotation "
                    f"of {theta} degrees")
            rotated_displacements = []
            for x, y, z in displacements:
                u = x*np.cos(np.deg2rad(theta))
                v = y*np.sin(np.deg2rad(theta))
                rotated_displacements.append(
                    [u, v, z]
                )
            displacements = rotated_displacements

        for iteration, displacement in zip(range(self.iterations), displacements):
            if self.print_info:
                print(f"Computing iteration {iteration}")

            displacement = random.choice(displacements)
            panel_xc_min = model_grid_def.xcmin + displacement[0]
            panel_usize_proportion = model_grid_def.usize/self.panel_usize
            panel_nu = np.ceil(panel_usize_proportion * model_grid_def.nu) + 1
            panel_yc_min = model_grid_def.ycmin + displacement[1]
            panel_vsize_proportion = model_grid_def.vsize/self.panel_vsize
            panel_nv = np.ceil(panel_vsize_proportion * model_grid_def.nv) + 1
            panel_zc_min = model_grid_def.zcmin + displacement[2]
            panel_zsize_proportion = model_grid_def.zsize/self.panel_zsize
            panel_nz = np.ceil(panel_zsize_proportion * model_grid_def.nz) + 1

            panel_grid_def_parameters = dict(
                xcmin = panel_xc_min,
                nu = panel_nu,
                usize = self.panel_usize,
                ycmin = panel_yc_min,
                nv = panel_nv,
                vsize = self.panel_vsize,
                zcmin = panel_zc_min,
                nz = panel_nz,
                zsize = self.panel_zsize,
                azimuth_deg = model_grid_def.azimuth_deg)

            panel_grid_def = rmsp.GridDef(**panel_grid_def_parameters, blockindices='full')
            panel_grid_data = rmsp.GridData(griddef=panel_grid_def)
            panel_grid_data['index'] = panel_grid_data.index

            xc, yc, zc = panel_grid_data.coords()
            panel_grid_data.loc[:, 'xc'] = xc
            panel_grid_data.loc[:, 'yc'] = yc
            panel_grid_data.loc[:, 'zc'] = zc

            self.panel_grid_data = panel_grid_data

            self.composites['panel_index'] = panel_grid_data.get_values_at(
                self.composites,
                columns_to_flag='index')['index']

            for variable in self.variables:
                if self.print_info:
                    print(f'\tComputing variable {variable}')
                for domain in self.domains:

                    if domain is not None:
                        if self.print_info:
                            print(f'\t\tComputing domain {domain}')
                        domain_filter = f'{self.domain_column} == {domain}'
                        grid_data_in_cudom = self.grid_data.query(domain_filter)
                        composites_dm_in_cudom = self.composites.query(domain_filter)
                    else:
                        if self.print_info:
                            print(f'\t\tComputing without domain filters')
                        domain = 0
                        grid_data_in_cudom = self.grid_data
                        composites_dm_in_cudom = self.composites


                    grdavg = rmsp.GridAverage(grid_data_in_cudom, panel_grid_def).build()
                    panel_grid_data_in_cudom = panel_grid_data.copy()
                    panel_grid_data_in_cudom['blocks_average'] = grdavg.apply(grid_data_in_cudom[variable])
                    panel_grid_data_in_cudom['blocks_filled'] = grdavg.get_target_filled()


                    declus = rmsp.DeclusterCell(only2d=True)
                    composites_dm_in_cudom.loc[:,'decluster_weight'] = declus.decluster(
                        composites_dm_in_cudom, variable, 150)

                    composites_in_panel = composites_dm_in_cudom.groupby(
                        'panel_index').size()
                    panel_grid_data_in_cudom.loc[
                        composites_in_panel.index,
                        'composites_in_panel'] = composites_in_panel

                    compute_weight = lambda x: (x['decluster_weight'].dot(x[variable]) ) / (x['decluster_weight'].sum())
                    weighted_measurement = composites_dm_in_cudom.groupby(
                        'panel_index').apply(compute_weight)

                    panel_grid_data_in_cudom.loc[weighted_measurement.index, 'weighted_measurement'] = weighted_measurement

                    panel_grid_data_in_cudom['iteration'] = iteration
                    panel_grid_data_in_cudom['variable'] = variable
                    panel_grid_data_in_cudom['domain'] = domain
                    panel_grid_data_in_cudom['displacements'] = str(displacement)

                    self.collected_data = self.collected_data.append(
                        panel_grid_data_in_cudom.dropna(),
                        ignore_index=True)

    def plot_scattered_data(
        self,
        filter=None,
        proportional_errors=[0.15, 0.20, 0.25],
        kernel_density=True,
        split_by_proportional_errors=False,
        figsize=(10,5)):

        if self.print_info:
            print(f"Assembling figures")
        if filter:
            valid_collected_data = self.collected_data.query(filter)
        else:
            valid_collected_data = self.collected_data

        all_figures = []

        for domain, data_domain in valid_collected_data.groupby('domain', dropna=False):
            if self.print_info:
                print(f"\tAssembling figure of domain {domain}")

            for variable, data_domain_var in data_domain.groupby('variable'):
                if self.print_info:
                    print(f"\t\tAssembling figure of variable {variable}")

                y = data_domain_var['blocks_average']
                x = data_domain_var['weighted_measurement']
                if not x.empty and not y.empty:
                    fig, (plot_axes, table_axes) = plt.subplots(2, 1, figsize=figsize)

                    plot_axes.set_title(f"Domain: {domain}, Variable: {variable}")

                    # Draw scatter points with kernel density
                    if kernel_density:
                        xy = np.vstack([x,y])
                        if len(x) > 5:
                            z = gaussian_kde(xy)(xy)
                            plot_axes.scatter(x, y, c=z)
                        else:
                            plot_axes.scatter(x, y)

                    # Draw scatter points splitted by proportional errors
                    mean_error = np.abs(x-y)/np.mean([x,y])
                    if split_by_proportional_errors:
                        labels = np.full(x.shape, np.nan)
                        for inferior_boundary, top_boundary in zip([0]+proportional_errors, proportional_errors+[np.inf] ):
                            gt_ix = mean_error >= inferior_boundary
                            lt_ix = mean_error < top_boundary
                            labels[np.logical_and(gt_ix, lt_ix)] = top_boundary
                        for g in np.unique(labels):
                            ix = np.where(labels == g)
                            plot_axes.scatter(x.iloc[ix], y.iloc[ix], label=g)
                        plot_axes.legend()

                    limit = max(max(x),max(y))
                    line_coords = np.array([0,limit])
                    plot_axes.plot(line_coords, line_coords, color='black')
                    for error in proportional_errors:
                        gamma = (error+2)/(2-error)
                        plot_axes.plot(line_coords, gamma*line_coords, color='gray', linestyle='dashed')
                        plot_axes.plot(line_coords, 1/gamma*line_coords, color='gray', linestyle='dashed')
                    plot_range = (0-0.05*limit, 1.05*limit)
                    plot_axes.set(xlim=plot_range, ylim=plot_range)
                    plot_axes.set_aspect(1)
                    plot_axes.figure.set_size_inches(figsize[0], figsize[0])

                    plot_axes.set_ylabel('blocks_average')
                    plot_axes.set_xlabel('weighted_measurement')


                    # Table assembly
                    total_data = len(x)
                    under_errors = []
                    proportions = []
                    header_columns = []
                    for error in proportional_errors:
                        under = np.sum(mean_error <= error)
                        under_errors.append(str(under))
                        proportions.append('%.2f' % (under/total_data*100)+'%')
                        header_columns.append('%.0f' % (error*100)+'%')

                    header = ['ERROR'] + header_columns
                    first_row = ['N TOTAL'] + [str(total_data)]*len(proportional_errors)
                    second_row = ['N DATOS BAJO ERROR'] + under_errors
                    third_row = ['% BAJO ERROR'] + proportions
                    table = [first_row, second_row, third_row]

                    table = table_axes.table(
                        cellText=table,
                        loc='top',
                        colLabels=header)
                        #rowLabels=rows)

                    table_axes.axis('off')
                    table_axes.axis('tight')
                    col_indexes = [-1] + list(range(len(header)))
                    table.auto_set_column_width(col=col_indexes)
                    fig. tight_layout(pad=3.0)

                    all_figures.append({
                        'domain': domain,
                        'variable': variable,
                        'figure': fig,
                        'plot_axes': plot_axes,
                        'table_axes': table_axes
                    })

                    fig.show()

        return all_figures
