from bokeh.models.markers import X
import numpy as np

from bokeh.plotting import figure
from bokeh.models import ColumnDataSource
from bokeh.transform import linear_cmap
from bokeh.palettes import Spectral6


def bokeh_section_plot(
    data_frame,
    var=None,
    solid=None,
    x_column='x',
    y_column='y',
    z_column='z',
    usize='usize',
    vsize='vsize',
    zsize='zsize',
    orient='xy',
    match_aspect=True,
    plot_width=500,
    plot_height=500,
    tools="pan,wheel_zoom,box_zoom,box_select,lasso_select,reset,save",
    coordinates=None,
    block_shrink=1,
    figures=None,
    block_outline=False,
    draw_centroids=False):

    if var:
        mapper = linear_cmap(
            field_name=var,
            palette=Spectral6,
            low=min(data_frame[var]),
            high=max(data_frame[var]))


    if block_shrink != 1:
        data_frame[usize] = block_shrink * data_frame[usize]
        data_frame[vsize] = block_shrink * data_frame[vsize]
        data_frame[zsize] = block_shrink * data_frame[zsize]

    if orient == 'xy':
        x_axis_label = "Easting (m)"
        y_axis_label = "Northing (m)"
        x_axis = x_column
        y_axis = y_column
        x_size = usize
        y_size = vsize
        ortogonal_axis = z_column
        figure_title = "Plan-section at {:.1f} m. Elev"
        orthogonal_coordinates = data_frame.loc[:, [ortogonal_axis]]

    if not coordinates:
        #print('No input coordinates passed')
        coordinates = orthogonal_coordinates.quantile(
            q=[0.25,0.5,0.75,1],
            interpolation='lower').drop_duplicates()
        coordinates = list(coordinates[ortogonal_axis])
        #print(f'Using coordinates {coordinates}')
    else:
        #print('Rounding input coordinates')
        rounded_coordinates = []
        for coordinate in coordinates:
            nearest_coordinate = orthogonal_coordinates.iloc[
                (orthogonal_coordinates[ortogonal_axis]-coordinate).abs().argsort()
                ].iat[0,0]
            rounded_coordinates.append(
                nearest_coordinate
            )
        coordinates = rounded_coordinates
        #print(f'Using coordinates {coordinates}')

    if not figures:
        figures = []
        for ortgonal_coord in coordinates:
            figures.append(
                figure(
                    title=figure_title.format(ortgonal_coord),
                    match_aspect=match_aspect,
                    plot_width=plot_width,
                    plot_height=plot_height,
                    tools=tools,
                    x_axis_label=x_axis_label,
                    y_axis_label=y_axis_label)
            )

    for ortgonal_coord, fig in zip(coordinates, figures):
        data_frame_at_level = data_frame.query(
            f'{ortogonal_axis}=={ortgonal_coord}')

        x_min = data_frame_at_level[x_axis].min()
        x_max = data_frame_at_level[x_axis].max()
        y_min = data_frame_at_level[y_axis].min()
        y_max = data_frame_at_level[y_axis].max()

        fig.x_range.start = x_min
        fig.x_range.end = x_max
        fig.y_range.start = y_min
        fig.y_range.end = y_max

        if data_frame.empty:
            print(f'No centroids at level {ortgonal_coord}')

        source = ColumnDataSource(data_frame_at_level)

        rect_parameters = {
            'x': x_axis,
            'y': y_axis,
            'height': y_size,
            'width': x_size,
            'source': source,
        }

        if block_outline:
            rect_parameters['fill_alpha'] = 0

        if var:
            rect_parameters['color'] = mapper

        fig.rect(**rect_parameters)

        if draw_centroids:
            circle_dot_parameters = {
            'x': x_axis,
            'y': y_axis,
            'size': 1,
            'source': source,
            'fill_color': None,
            }

            if var:
                 circle_dot_parameters['color'] = mapper

            fig.circle_dot(**circle_dot_parameters)


        fig.left[0].formatter.use_scientific = False
        fig.xaxis[0].formatter.use_scientific = False

        if solid:
            solid_slice = solid.get_slice(
                orient,
                ortgonal_coord,
                tolerances=0.01)

            x_coords = solid_slice.x
            y_coords = solid_slice.y
            z_coord = solid_slice.z

            if orient=='xy':
                fig.multi_line(
                    x_coords[solid_slice.tri_indices].tolist(),
                    y_coords[solid_slice.tri_indices].tolist(),
                    color='black'
                )

    return figures


from bokeh.models import Spinner
from bokeh.layouts import column, row
from bokeh.models import CustomJS, RangeSlider
from bokeh.layouts import gridplot

def bokeh_cross_section(
        data_frame, **kwargs
    ):

    figures = bokeh_section_plot(
        data_frame=data_frame,
        **kwargs
    )

    # Retrieve the biggest figure range for all the figures
    min_x = min_y = np.inf
    max_x = max_y = -np.inf
    for figure in figures:
        fig_min_y = figure.y_range.start
        fig_max_y = figure.y_range.end
        fig_min_x = figure.x_range.start
        fig_max_x = figure.x_range.end

        if fig_min_y < min_y:
            min_y = fig_min_y
        if fig_max_y > max_y:
            max_y = fig_max_y
        if fig_min_x < min_x:
            min_x = fig_min_x
        if fig_max_x > max_x:
            max_x = fig_max_x

    # Link all figures ranges to the first one
    p = figures[0]
    for figure in figures[1:]:
        figure.x_range = p.x_range
        figure.y_range = p.y_range

    range_x_slider = RangeSlider(
        title=" Adjust X-Axis range",
        start=min_x,
        end=max_x,
        step=0.1,
        value=(p.x_range.start, p.x_range.end+1),
    )

    range_x_slider.js_link("value", p.x_range, "start", attr_selector=0)
    range_x_slider.js_link("value", p.x_range, "end", attr_selector=1)

    range_y_slider = RangeSlider(
        title=" Adjust Y-Axis range",
        start=p.y_range.start,
        end= p.y_range.end,
        step=0.1,
        value=(p.y_range.start, p.y_range.end+1),
    )

    range_y_slider.js_link("value", p.y_range, "start", attr_selector=0)
    range_y_slider.js_link("value", p.y_range, "end", attr_selector=1)

    selectors = column([range_x_slider, range_y_slider], width=500)
    return column(selectors, gridplot(figures, ncols=2))



from bokeh.layouts import column
from bokeh.models import ColumnDataSource, Slider, CustomJS, Select

from bokeh.plotting import figure, output_file, show
import numpy as np
def z_slider(data_frame, **kwargs):

    data_frame['x_original'] = data_frame['x']
    data_frame['y_original'] = data_frame['y']
    data_frame['z_original'] = data_frame['z']

    source = ColumnDataSource(data_frame)

    # Create plots and widgets
    plot = figure()
    plot.x_range.start = data_frame['x_original'].min()
    plot.x_range.end = data_frame['x_original'].max()
    plot.y_range.start = data_frame['y_original'].min()
    plot.y_range.end = data_frame['y_original'].max()

    plot.rect('x', 'y', width='usize', height='vsize', source=source)

    # Create Slider object
    z_coords = data_frame['z'].unique()
    min_zcoord = min(z_coords)
    z_coords = [str(z) for z in z_coords]
    slider = Select(
        options=z_coords,
        value=str(min_zcoord),
        title='Z coordinate')

    # Adding callback code
    callback = CustomJS(args=dict(source=source, val=slider),
                        code="""
        const data = source.data;
        const z_level = val.value;
        const x_original = data['x_original'];
        const y_original = data['y_original'];
        const z_original = data['z_original'];
        const x = data['x'];
        const y = data['y'];
        for (var i = 0; i < z_original.length; i++) {
           if(z_original[i] != z_level){
               x[i] = 0;
               y[i] = 0;
           } else{
                x[i] = x_original[i];
                y[i] = y_original[i];
           }
        }
        x.filter(e => e !== 0);
        y.filter(e => e !== 0);
        source.change.emit();
        console.log(z_level);
    """)

    slider.js_on_change('value', callback)

    # Arrange plots and widgets in layouts
    layout = column(slider, plot)


    return(layout)
