import json
import os
from subprocess import PIPE, run


class JupyterRenderException(Exception):
    pass


class JupyterRender(object):
    """ A class to render or translate content in code blocks and
    markdown blocks of a Jupyter notebook."""


    def __init__(self,
        template_file_name,
        code_blocks_translation=dict(),
        markdown_blocks_translation=dict(),
        translated_notebook_file_name=None,
        keep_translated_file=False,
        print_info=False):

        self.print_info = print_info

        if not isinstance(code_blocks_translation, dict):
            error_message = "The code block translation has to be a dictionary"
            raise JupyterRenderException(error_message)

        self.code_blocks_translation = code_blocks_translation

        if not isinstance(markdown_blocks_translation, dict):
            error_message = "The markdown block translation has to be a dictionary"
            raise JupyterRenderException(error_message)

        self.markdown_blocks_translation = markdown_blocks_translation

        # Check template_file_name is a valid file
        if not self.check_file_exists(template_file_name):
            error_message = f"File {template_file_name} can't be found"
            raise JupyterRenderException(error_message)

        self.template_file_name = template_file_name

        # Retrieve file content
        self.template_notebook_content = \
            self.get_notebook_file_content(template_file_name)

        # Set translated notebook file name
        if not translated_notebook_file_name:
            self.translated_notebook_file_name = \
                os.path.splitext(template_file_name)[0] \
                    + '-translated.ipynb'
        else:
            self.translated_notebook_file_name = translated_notebook_file_name

        # Set keep_translated_file flag
        if not isinstance(keep_translated_file, bool):
            error_message = f"The kee_translated_file flag need to be a bool not {keep_translated_file} "
            raise JupyterRenderException(error_message)
        else:
            self.keep_translated_file = keep_translated_file

        if self.print_info:
            print("Setted up a render translation of Jupyter notebook content")
            print(f"\tOriginal file {self.template_file_name}")
            print(f"\tObjective file {self.translated_notebook_file_name}")
            print(f"\tCode blocks translation:")
            for original, translated in self.code_blocks_translation.items():
                print(f"\t\t{original} -> {translated}")
            print(f"\tMarkdown blocks translation:")
            for original, translated in self.markdown_blocks_translation.items():
                print(f"\t\t{original} -> {translated}")


    def check_file_exists(self, file):
        return os.path.exists(file)


    def get_notebook_file_content(self, file):
        with open(self.template_file_name) as f:
            lines = f.readlines()

        with open(self.template_file_name) as notebook_file:
            notebook_content = json.load(notebook_file)

        return notebook_content

    def apply_translation(self):

        translated_notebook_content = self.template_notebook_content.copy()
        for cell in translated_notebook_content['cells']:
            if cell['cell_type'] == 'markdown':
                for index in range(len(cell['source'])):
                    markdown_line=cell['source'][index]
                    for translation_key, translation in self.markdown_blocks_translation.items():
                        markdown_line = markdown_line.replace(translation_key, translation)
                    cell['source'][index] = markdown_line


            elif cell['cell_type'] == 'code':
                for index in range(len(cell['source'])):
                    code_line = cell['source'][index]
                    for translation_key, translation in self.code_blocks_translation.items():
                        code_line = code_line.replace(translation_key, translation)
                    cell['source'][index] = code_line

        translated_notebook = json.dumps(
            translated_notebook_content,
            indent=4,
            sort_keys=True)

        with open(self.translated_notebook_file_name, "w") as notebook_output_file:
            notebook_output_file.write(translated_notebook)

        if self.print_info:
            print(f"\tTranslated file created {self.translated_notebook_file_name}")

        return self.translated_notebook_file_name


    def execute_translated_notebook(self, execution_option='notebook', output_dir=None):

        # Verify execution_option parameter
        execution_options = ['notebook', 'html_embed']
        if not execution_option in execution_options:
            error_message = f"The execution_option {execution_option} needs to be one of {execution_options}"
            raise JupyterRenderException(error_message)

        # Verify output directory
        if output_dir:
            if not os.path.isdir(output_dir):
                os.makedirs(output_dir)
                if self.print_info:
                    print(f"\t Creating output folder {output_dir}")

        # Check if translated file exists, if not execute translation
        traslated_file_exits = os.path.exists(self.translated_notebook_file_name)
        if not traslated_file_exits:
            print(
                f"\t Translated file {self.translated_notebook_file_name} "
                "not found, applying translation.")
            self.apply_translation()

        if execution_option == 'notebook':
            jupyter_execute_command = \
                f"jupyter nbconvert --to notebook " \
                f"--ExecutePreprocessor.timeout=999999 " \
                f"--template toc2 " \
                f"--inplace "\
                f"--execute {self.translated_notebook_file_name}"

        elif execution_option == 'html_embed':
            jupyter_execute_command = \
                f"jupyter nbconvert --to html_embed " \
                f"--ExecutePreprocessor.timeout=999999 " \
                f"--template toc2 " \
                f"--execute {self.translated_notebook_file_name}"

        if output_dir:
            jupyter_execute_command = jupyter_execute_command + \
                f" --output-dir='{output_dir}' "

        if self.print_info:
            print(f"\t Executing: {jupyter_execute_command}")

        result = run(
            jupyter_execute_command,
            stdout=PIPE,
            stderr=PIPE,
            universal_newlines=True)

        if self.print_info:
            if result.returncode == 0:
                print('\t Translation execution completed')
            else:
                print('\t Translation bad executed')

            print('\t Jupyter nbconvert execution returns: ')
            print(result.stdout)
            print(result.stderr)

        if not self.keep_translated_file:
            os.remove(self.translated_notebook_file_name)

        return result.returncode
