import os
import subprocess
from pathlib import Path
import json
from slugify import slugify
import pandas as pd
from openpyxl import load_workbook


class SnippetLoaderlException(Exception):
    pass


class SnippetLoader():

    default_snippets = {
        'numpy' : 0,
        'scipy' : 1,
        'matplotlib': 2,
        'sympy': 3,
        'pandas': 4,
        'astropy': 5,
        'h5py': 6,
        'numba': 7,
        'python': 8,
        'markdown': 9
    }

    def __init__(
        self,
        snippets_file='my_snippets.xlsx',
        default_snippets_to_remove=None):

        if not os.path.exists(snippets_file):
            raise SnippetLoaderlException(
                f"Snippets file {snippets_file} not found"
            )

        snippets_wb = load_workbook(
            snippets_file,
            read_only=True,
            keep_links=False)

        snippets_menu = dict()
        for sheet in snippets_wb.sheetnames:
            snippets_data = pd.read_excel(
                snippets_file,
                sheet_name=sheet)

            for expected_column in ['menu_level', 'name', 'snippet']:
                if not expected_column in snippets_data.columns:
                    raise SnippetLoaderlException(
                        f"column {expected_column} not found on sheet {sheet}")

            snippets_menu[sheet] = snippets_data

        self.snippets_menu_names = snippets_wb.sheetnames # Each sheet creates a root menu
        self.snippets_menu = snippets_menu  # Data to be loaded to custom file


        get_jupyter_config_dir_command = "jupyter --config-dir"
        jupyter_cofig_dir = subprocess.check_output(
            get_jupyter_config_dir_command,
            shell=True).strip().decode()

        custom_file = os.path.join(jupyter_cofig_dir, 'custom', 'custom.js')
        custom_folder = os.path.dirname(custom_file)
        custom_folder_exists = os.path.exists(custom_folder)
        custom_file_exists = os.path.exists(custom_file)

        if not custom_folder_exists:
            print(f'Creating custom folder {custom_folder}')
            os.makedirs(custom_folder)

        if not custom_file_exists:
            print(f'Creating custom file {custom_file}')
            with open(custom_file, "w") as f:
                f.write("created")

        self.custom_file = custom_file # Custom  file valid path

        if isinstance(default_snippets_to_remove, list):
            for snippet in default_snippets_to_remove:
                if not snippet in self.default_snippets.keys():
                    error_message = f"Default snippet {snippet} doesn't exists"
                    raise SnippetLoaderlException(error_message)

            default_snippets_to_remove = {
                name:self.default_snippets[name]
                for name in default_snippets_to_remove if name in self.default_snippets
                }
        elif default_snippets_to_remove is not None:
            error_message = f"The parameter default_snippets_to_remove is eithr None or a list"
            raise SnippetLoaderlException(error_message)

        self.default_snippets_to_remove = default_snippets_to_remove

    def update_snippets_file(self):
        print(f'Creating new content on file {self.custom_file}')
        custom_file_content = self.assembly_custom_file_js_content()
        print(f'Updating file {self.custom_file}')
        with open(self.custom_file, "w") as custom_file_opened:
            custom_file_opened.write(custom_file_content)
        print(f'Done, remember to close and reload your Jupyter notebook server')

    def assembly_custom_file_js_content(self):

        custom_file_content =     """
            require(["nbextensions/snippets_menu/main"], function (snippets_menu) {
            console.log('Loading `snippets_menu` customizations from `custom.js`');
            var horizontal_line = '---';
        """

        json_menues = self.assembly_snippets_menues()
        for var in json_menues.keys():
            menu_variable = "var "+ slugify(var) + " = " + json.dumps(
                json_menues[var],  indent=4)+ "; \n"
            custom_file_content = custom_file_content + menu_variable


        custom_file_content = custom_file_content + """
                snippets_menu.options['menus'] = snippets_menu.default_menus;
                snippets_menu.options['menus'][0]['sub-menu'].push(horizontal_line);
            """

        for var in json_menues.keys():
            push_command = f"snippets_menu.options['menus'][0]['sub-menu'].push({slugify(var)});"
            custom_file_content = custom_file_content + push_command

        if self.default_snippets_to_remove:
            for name in sorted(
                self.default_snippets_to_remove,
                key=self.default_snippets_to_remove.get,
                reverse=True):

                snippet_menu_number = self.default_snippets_to_remove[name]
                custom_file_content = custom_file_content + f"""
                    snippets_menu.default_menus[0]['sub-menu'].splice({snippet_menu_number}, 1);
                """
                print(f"Removing default snippet {name}")

        custom_file_content = custom_file_content + """
                console.log('Loaded `snippets_menu` customizations from `custom.js`');
            });
            """

        return custom_file_content

    def assembly_snippets_menues(self):
        json_menues = {}
        for sheet in self.snippets_menu_names:
            print(f'Creating menu {sheet}')
            menu = {
                'name': sheet,
                'sub-menu': []}
            first_snippets = self.snippets_menu[sheet].query('menu_level.isnull()')
            first_snippets = first_snippets.sort_values('name')
            for _, snippet_data in first_snippets.iterrows():
                name = snippet_data['name']
                snippet = snippet_data['snippet'].split('\n')
                menu['sub-menu'].append({
                    'name': name,
                    'snippet': snippet
                    })
                print(f"\tCreating snippet {name}")

            last_snippets = self.snippets_menu[sheet].dropna()

            if not last_snippets.empty:
                menu['sub-menu'].append('---')

                for submenu_name, submenu_snippet_data in last_snippets.groupby('menu_level'):
                    print(f'\tCreating sub-menu {submenu_name}')
                    sub_menu = []
                    for _, snippet_data in submenu_snippet_data.iterrows():
                        name = snippet_data['name']
                        snippet = snippet_data['snippet'].split('\n')
                        sub_menu.append({
                            'name': name,
                            'snippet': snippet
                            })
                        print(f"\t\tCreating snippet {name}")

                    menu['sub-menu'].append({
                        'name': submenu_name,
                        'sub-menu': sub_menu
                    })

            json_menues[sheet]= menu

        return json_menues
