import os
import zipfile
import string
import random


class CompressedFileHandlerException(Exception):
    pass


class CompressedFileHandler(object):
    """
    A class to manage compressed files on a Python context
    manager"""
    def __init__(self,
                 file_name,
                 compressed_format='zip',
                 keep_temporal_files=False,
                 force_unique_files=True):
        self.file_name = file_name
        self.compressed_format = compressed_format
        self.keep_temporal_files = keep_temporal_files
        self.force_unique_files = force_unique_files

    def __enter__(self):
        self.check_file_exists(
            file_name=self.file_name)
        self.get_objective_folder(
            file_name=self.file_name)

        file_handled_compression_format = \
            self.get_compression_format(self.file_name)

        self.temporal_file_names  = []
        if file_handled_compression_format == 'zip':
            zip_file = zipfile.ZipFile(self.file_name, 'r')
            zipinfos = zip_file.infolist()
            for zipinfo in zipinfos:
                temporal_file_name = self.get_temporal_name(
                    zipinfo.filename,
                    keep_extension=True)
                temporal_file_name = os.path.join(
                    temporal_file_name
                    )

                zipinfo.filename = temporal_file_name
                unzipped_file = zip_file.extract(
                    zipinfo,
                    path=self.objective_folder)

                self.temporal_file_names.append(unzipped_file)
        else:
            error_message = f"The file {self.file_name} isn't a compressed one"
            raise CompressedFileHandlerException(error_message)

        if len(self.temporal_file_names) == 1:
            return self.temporal_file_names[0]
        else:
            return self.temporal_file_names

    def get_compression_format(self, file):
        if zipfile.is_zipfile(file):
            return 'zip'
        return None

    def get_temporal_name(self, file_name, keep_extension=False):
        chars=string.ascii_uppercase + string.digits

        file_name = os.path.basename(file_name)
        file_name_without_extension = os.path.splitext(file_name)[0]
        file_extension =  os.path.splitext(file_name)[1]

        if self.force_unique_files:
            temporal_file_without_extension = \
                file_name_without_extension + \
                ''.join(random.sample(chars, 6))
        else:
            temporal_file_without_extension = file_name_without_extension

        if keep_extension:
            return temporal_file_without_extension + \
                file_extension
        else:
            return temporal_file_without_extension + \
                '.' + self.compressed_format

    def check_file_exists(self, file_name):
        if not os.path.exists(file_name):
            error_message = f"File {file_name} doesn't exists"
            raise CompressedFileHandlerException(error_message)

    def get_objective_folder(self, file_name):
        self.objective_folder = os.path.dirname(file_name)

    def __exit__(self, exception_type, exception_value, traceback):
        # If execution fails raise False flag and keep temporal files
        if exception_type:
            return False

        if not self.keep_temporal_files:
            for file in self.temporal_file_names:
                os.remove(file)

        return True




class UncompressedFileHandlerException(Exception):
    pass


class UncompressedFileHandler(object):
    """
    A class to manage uncompressed files on a Python context
    manager"""
    def __init__(self,
                 file_name=None,
                 files_names=None,
                 output_file_name=None,
                 compressed_format='zip',
                 keep_original_files=False,
                 force_unique_file=True):

        self.compressed_format = compressed_format
        self.keep_original_files = keep_original_files
        self.force_unique_file = force_unique_file

        # Check incoming files
        if file_name:
            if not isinstance(file_name, str):
                error_message = \
                    f"The file {file_name} has to be a string path to file"
                raise UncompressedFileHandlerException(error_message)

            if files_names:
                if not isinstance(files_names, list):
                    error_message = \
                        f"The files {files_names} has to be a string path to file"
                    raise UncompressedFileHandlerException(error_message)
                files_names = [file_name] + files_names
            else:
                files_names = [file_name]

        if not files_names or not isinstance(files_names, list):
            error_message = 'A list of files should be provided'
            raise UncompressedFileHandlerException(error_message)
        self.files_names = files_names

        if not output_file_name:
            self.output_file_name = self.get_temporal_name(
                files_names[0])
        else:
            self.output_file_name = output_file_name


    def __enter__(self):
        for file in self.files_names:
            self.check_file_exists(
                file_name=file)


        if self.compressed_format == 'zip':
            self.file = zipfile.ZipFile(
                self.output_file_name,
                'w',
                zipfile.ZIP_DEFLATED)

            for file in self.files_names:
                self.file.write(file)

        else:
            error_message = "Only zip compression is managed by now."
            raise UncompressedFileHandlerException(error_message)

        return self.output_file_name



    def get_temporal_name(self, file_name, keep_extension=False):
        chars=string.ascii_uppercase + string.digits
        file_name_without_extension = os.path.splitext(file_name)[0]
        file_extension =  os.path.splitext(file_name)[1]

        if self.force_unique_file:
            temporal_file_without_extension = \
                file_name_without_extension + \
                ''.join(random.sample(chars, 6))
        else:
            temporal_file_without_extension = file_name_without_extension

        if keep_extension:
            return temporal_file_without_extension + \
                file_extension
        else:
            return temporal_file_without_extension + \
                '.' + self.compressed_format

    def check_file_exists(self, file_name):
        if not os.path.exists(file_name):
            error_message = f"File {file_name} doesn't exists"
            raise UncompressedFileHandlerException(error_message)


    def __exit__(self, exception_type, exception_value, traceback):
        # If execution fails raise False flag and keep temporal files
        if exception_type:
            return False

        self.file.close()

        if not self.keep_original_files:
            for file in self.files_names:
                os.remove(file)

        return True
